#!/usr/bin/perl

# -------------  script for computing the rate of star captures by a supermassive black hole -----------
# the galaxy model is provided by a text file with various parameters as functions of radius (mass, distribution function, and diffusion coefficients), which is created by an auxiliary program `mkspher'; 
# the model is scaled to a real galaxy by specifying the number of stars and the scale radius of tab file [in pc], and possibly the radius of tidal disruption in pc (otherwise it is assumed to be 4x Schwarzschild radius).
# output consists of a number of computed quantities, most importantly, total captured mass during the given time, and several other parameters that are appended to a text file $summary.
$summary="summary.txt";
foreach $arg (@ARGV)    # list of possible command-line arguments
{
  $used=0;
  if($arg=~/tab=(.+)/)                { $used=1; $tab=$1;    }  # name of tab file from 'mkspher' (smooth-potentail properties) [no default]
  if($arg=~/mass=([\d\.\+\-de]+)/i)   { $used=1; $Nbody=$1;  }  # total mass of galaxy in solar masses; the same quantity gives the number of stars in galaxy, assuming solar-mass stars and neglecting the contribution of Mbh to the total mass [default: read from last line of the tab file]
  if($arg=~/time=([\d\.\+\-de]+)/i)   { $used=1; $timeyr=$1; }  # galaxy lifetime, in years [10^10 by default]
  if($arg=~/rscale=([\d\.\+\-de]+)/i) { $used=1; $rscale=$1; }  # scale radius of tab file in parsecs [default: 1pc]
  if($arg=~/rcapt=([\d\.\+\-de]+)/i)  { $used=1; $rcaptp=$1; }  # capture radius in parsecs [by default 8Mbh/c^2, the periapsis radius of direct capture]
  if($arg=~/Rsep=([\d\.\+\-de]+)/i)   { $used=1; $Rsep0=$1;  }  # R_sep is the parameter of flattening for axisymmetric models [0 by default, means spherical case]; 1-Rsep roughly equals the axis ratio z/x
  if($arg=~/Fdrain=([\d\.\+\-de]+)/i) { $used=1; $Fdrain=$1; }  # fractional initial population of stars in the loss region [0 by default, means it was empty; 1 means isotropic distribution, i.e. no depletion of initial loss region]
  if($arg=~/triax/i)                  { $used=1; $triax=1;   }  # triaxial loss region instead of axisymmetric (only has effect if R_sep>0)
  if($arg=~/nbunits/i)                { $used=1; $nbunits=1; }  # use N-body units instead of physical units
  if($arg=~/multdc=([\d\.\+\-de]+)/i) { $used=1; $multdc=$1; }  # tweak the value of diffusion coefficient multiplier in the tab file [by default =(ln N)/N]
  if(!$used) { die "Parameter $arg unknown"; }
}
$output=0;  # set to 1 to get a detailed text file output (undocumented)

# intermezzo: two convenience functions
sub pow
{
  $x=shift;
  $p=shift;
  return exp($p*log($x));
}
sub pow2
{
  $x=shift;
  return $x*$x;
}
if(!$tab)    { die "Must specify input table: tab=???"; }
if(!$timeyr) { $timeyr=1e10; }
if(!$Rsep0)  { $Rsep0=0; }
if(!$Fdrain) { $Fdrain=1; }

open FI, "<$tab" or die("Can't open input file $tab");
<FI>;  #skip first line
while(<FI>)
{
  chomp;
  @L=split(/\s+/);
  if($L[0]==0) { $Mbh=$L[1]; } else { push @Energy, $L[2]; }
  $Mtot=$L[1];
}
close FI;

if(!$Mbh) { die "No black hole is found at r=0 in the tab file $tab"; }
if(!$Nbody) { 
  if($Mtot < 1e4 && !$nbunits) { die "No mass scale (mass=???) is provided and masses in the tab file apparently are not in units of solar mass"; }
  $Nbody=($Mtot-$Mbh);  # otherwise assume that mass scale is in solar masses
}
$mscale=$Nbody/($Mtot-$Mbh); # mass scale: how many solar masses in one mass unit in the tab file
$Mbhsol=$Mbh*$mscale;        # black hole mass in solar masses
if(!$rscale) { $rscale=1; }  # radius scale: how many parsecs in one length unit in tab file
if(!$rcaptp) { $rcaptp=8*6.67e-8/pow2(3e10)*$Mbh*2e33*$mscale/3.08e18; }  # 4x Schwarzschild radius in units of parsecs
$rcapt=$rcaptp/$rscale;      # capture radius in length units of tab file

# figure out important quantities
$loglambda=log(0.3*$Nbody*$Mbh/($Mtot-$Mbh));      # Coulomb logarithm; factor 0.3 is quite arbitrary (see multdc=... for a tweak)
#$loglambda=log(0.05*$Nbody);
if(!$multdc) 
{ $multdc=$loglambda/$Nbody*($Mtot-$Mbh); }        # factor in the definition of diffusion coefficients which scales with the number of stars
if($nbunits) { $tscale=1; } else
{ $tscale=1.48e7*pow($rscale,1.5)/sqrt($mscale); } # model time unit in years
$maxtime=$timeyr/$tscale;                          # evolution time in model units
print "rscale=$rscale, mscale=$mscale, tscale=$tscale, multdc=$multdc\n";

# another quick scan through the tab file to find out influence radius and related stuff
open FI, "<$tab";
<FI>;
while(<FI>) {
  chomp;
  @L=split(/\s+/);
  if($L[1]>3*$Mbh && !$rinflpc) { # record influence radius and corresponding dynamical quantities
    $rinflpc=$L[0] * $rscale;     # radius in pc
    $tradrinfl=$L[7] * $tscale;   # dynamical time in yr
    $TrelSrinfl=0.34*pow($L[10],3)/$L[3]/$multdc * $tscale;  # relaxation time in the standard (local) definition, in yr
  }
}
close FI;
print "Nbody=$Nbody, Mbh=$Mbhsol M_sol, rcapt=$rcaptp pc, evol.time=$timeyr yr, LnLambda=$loglambda\n";
print "rinfl=$rinflpc pc, Tdyn(rinfl)=$tradrinfl yr, Trel(rinfl)local=$TrelSrinfl yr\n";

# compute theoretical capture rates and loss cone occupation from `mkspherical' tables
{
  open FI, "<$tab" or die("Can't open input file $tab");
  $tab=~s/\.tab//;
  $suffix=$Rsep0>0 ? ($triax? "tri":"axi") : "sph";
  $captout="${tab}_rc${rcapt}_${suffix}.capt";
  if($output) {
    open FO, ">$captout" or die("Can't open output file $captout");
  }
  $emptylc=$remptylc=$fulllc=$rfulllc=$fulllcaxi=$rfulllcaxi=$rinfl=0;
  <FI>;  #skip first line
  <FI>;  #skip Mbh at r=0
  $indx=0;
  while(<FI>)
  {
    chomp;
    @L=split(/\s+/);
    # attention to the column indices! they tend to change with each new version of mkspherical...
    my $colMofR= $L[1];  # cumulative mass M(r) (including black hole)
    my $colE   = $L[2];  # energy E (=potential at the given radius $L[0])
    my $colDF  = $L[4];  # value of distribution function f(E)
    my $colTrad= $L[7];  # radial period
    my $colRcir= $L[8];  # radius of a circular orbit with the energy given by $colE; roughly equal to $L[0]/2
    my $colLcir= $L[9];  # angular momentum of a circular orbit
    my $colDRR = $L[15]; # limiting value of diffusion coefficient [ lim D_RR/R (R->0) ]
    my $deltaE = $Energy[$indx+1] ? $indx>0 ? ($Energy[$indx+1]-$Energy[$indx-1])/2 : $Energy[1]-$Energy[0] : $Energy[$indx]-$Energy[$indx-1];
    my $mass=$colDF*4*pow2(3.14*$colLcir)*$colTrad * $deltaE * $mscale;
    $indx++;
    if($#L>3)
    {
      $Rsep= $colMofR>$Mbh && $Rsep0>0 ? $Rsep0*0.5*(1+($colMofR<7*$Mbh?pow(1-pow(($colMofR-$Mbh)/(6*$Mbh),2.0/3),1.5):0)):0;
      my $Trad=$colTrad;
      my $Rcapt=2*$Mbh*$rcapt/pow2($colLcir);
      if($Rcapt>0)
      {
        $FluxFullLC=$Rcapt/$Trad;
        $DRRoverR=$colDRR*$multdc;
        my $q=$DRRoverR*$Trad/$Rcapt;
        my $alpha=pow(pow($q,4)+pow2($q),0.25);
        my $denom=log(1/$Rcapt)-1; 
        if($denom>0) {
          $Reff= ($Rsep<$Rcapt) ? $Rcapt: $Rsep*(0.1+0.9*sqrt($Rcapt/$Rsep));
          #$Reff= ($Rsep<$Rcapt) ? $Rcapt: exp(-3.14/sqrt($Rsep)-1);  # Magorrian&Tremaine
          if($triax) { $Rlr=$Rsep0/4; }  # in the triaxial case, this gives the fraction of phase space occupied by centrophilic orbits
          else { $Rlr = ($Rsep<$Rcapt) ? $Rcapt: sqrt($Rsep*$Rcapt); } # volume of loss region (loss wedge or region of chaotic orbits, or just the loss cone in the spherical case)
          $denomeff=log(1/$Reff)-1+0*$Rsep/3;
          if($denomeff<0){$denomeff=0;}
          if($Rsep>0){
            $Tosc=$Trad*0.5*$Mbh/(($colMofR-$Mbh)*0.35) /0.8/sqrt($Rsep) * (1 + (($colMofR-$Mbh)/$Mbh)*0.5);
            $paxi=$Trad/$Tosc * $Rlr/$Rcapt; #*3.14/2*sqrt($Rsep/$Rcapt);
            #$alphaeff=$DRRoverR*$Tosc / $Rlr * ($paxi>1?$paxi:1);
            $alphaeff=$q * ($paxi<1?1/$paxi:1);
            if($alphaeff<$alpha) { $alphaeff=$alpha; }  #hmmm?
          } else { $alphaeff=$alpha; $paxi=0; $Tosc=0; }
          $FluxOverF=$DRRoverR/($denomeff+$alphaeff);
          $Flux = $FluxOverF * $mass;
          $TotalFlux+=$Flux*$maxtime;  # total number of captured stars in this energy interval during the entire evolution time (stationary case)
          if($alpha>1 && !$emptylc)   { $emptylc=-$colE; $remptylc=$colRcir; }  # record transition from "empty" to "intermediate" loss cone
          if($alpha>$denom && !$fulllc) { $fulllc=-$colE; $rfulllc=$colRcir; }  # record transition from "intermediate" to "full"  loss cone
          # time-dependent flux
          $tstep=$output ? $maxtime/100 : $maxtime;  # fine timestep if detailed output is required
          my $tdflux=`./dif1d dc=$DRRoverR tmax=$maxtime tstep=$tstep rmin=$Reff alpha=$alphaeff Rdrain=$Rlr MaxFlux=$FluxFullLC Fdrain=$Fdrain`;
          my @tdflux=split(/\n/,$tdflux);
          my $i=0;
          foreach $tflux(@tdflux)
          {
            @LIST=split(/\t/,$tflux);
            $FPtimes[$i]=$LIST[0];
            $FPflux      {$LIST[0]} += $mass*$LIST[1]/$tscale;
            $FPfluxTotal {$LIST[0]} += $mass*$LIST[2];
            $FPdrain     {$LIST[0]} += $mass*$LIST[3]/$tscale;
            $FPdrainTotal{$LIST[0]} += $mass*$LIST[4];
            $i++;
          }
          $TDflux=$mass*$LIST[1];      # instantaneous capture rate at the end of evolution time
          $TDcaptmass=$mass*$LIST[2];  # total number of captured stars by the end of evolution, from the time-dependent solution
          $TDdrain=$mass*$LIST[3];     # instantaneous drain rate at the end of evolution time
          $TDdrainmass=$mass*$LIST[4]; # total number of drained stars
          $TotalTDflux+=$TDcaptmass;
          $TotalDrain+=$TDdrainmass;
          $TodayTDflux+=$TDflux;
          $TodayTDdrain+=$TDdrain;
          my $Tdrain = $Trad*$Rlr/$Rcapt * ($paxi<1&&$paxi>0 ? 1/$paxi : 1);
          if($prevR>0 && $output)
          {
          print FO join("\t", $L[0], $colRcir, 
            $Flux*$maxtime/($colRcir-$prevR),
            $TDcaptmass/($colRcir-$prevR),
            $TDdrainmass/($colRcir-$prevR),
            $TDflux*$maxtime/($colRcir-$prevR),
            $TDdrain*$maxtime/($colRcir-$prevR),
            $Tdrain, $alpha, $Rsep, $colRcir-$prevR
            )."\n";
          }
        }
      }
    }
    $prevR=$colRcir;
  }
  close FI;
  if($output) {
    close FO;

    # write time-dependent flux
    $fluxout="${tab}_rc${rcapt}_${suffix}.flux";
    open FO, ">$fluxout" or die("Can't open output file $fluxout");
    foreach $FPtime(@FPtimes)
    {
      print FO join("\t", $FPtime*$tscale, $FPflux{$FPtime}, $FPfluxTotal{$FPtime}, $TotalFlux*$FPtime/$maxtime, 
        $FPdrain{$FPtime}, $FPdrainTotal{$FPtime})."\n";
    }
    close FO;
  }
  print "r_empty=".($remptylc*$rscale)." pc; r_full=".($rfulllc*$rscale)." pc\n";
  print "Captured mass[Msol]: steady-state=$TotalFlux, time-dep=$TotalTDflux, drain=$TotalDrain\n";
  print "Present-day capture rate[Msol/yr]: steady-state=".($TotalFlux/$timeyr).", time-dep=".($TodayTDflux/$tscale).", drain=".($TodayTDdrain/$tscale)."\n";
}
# write summary information to a file `summary.txt'
if(!-e $summary) { $hdr=1; }
open FO, ">>$summary";
if($hdr) { print FO join("\t", 
  "M_bh[Msol]",                # black hole mass in M_sol
  "r_lc[pc]",                  # loss cone radius in pc
  "time_evol[yr]",             # evolution time (galaxy age) in years
  "R_sep[dimless]",            # dimensionless coefficient related to the degree of flattening (0 for spherical case)
  "M_capt,steady-state[Msol]", # total captured mass by the end of evolution from the stationary expression for flux
  "M_capt,time-dep[Msol]",     # the same from time-dependent solution of diffusion equation
  "M_drain[Msol]",             # total captured mass from draining of the loss region
  "Mdot(t_evol)[Msol/yr]",     # capture rate (M_sol/year) at the end of evolution from the time-dependent solution
  "Mdot_drain[Msol/yr]"),       # the same only taking into account the draining
  "\n"; }
print FO join("\t", 
  $Mbh*$mscale,
  $rcaptp,
  $timeyr,
  $Rsep0,
  $TotalFlux,
  $TotalTDflux,
  $TotalDrain,
  $TodayTDflux/$tscale,
  $TodayTDdrain/$tscale
). "\n";
close FO;
