#pragma once
#include <string>
#include <algorithm>
#include <sstream>
#include <fstream> 
#include <map>

/**
* The StringVariant class is a simple string-based variant implementation that allows
* the user to easily convert between simple numeric/string types.
*/

class StringVariant
{
	std::string data;
public:
	StringVariant() : data() {}

	template<typename ValueType>
	StringVariant(ValueType val)
	{
		std::ostringstream stream;
		stream << val;
		data.assign(stream.str());
	}

	template<typename ValueType>
	StringVariant& operator=(const ValueType val)
	{
		std::ostringstream stream;
		stream << val;
		data.assign(stream.str());

		return *this;
	}

	template<typename NumberType>
	NumberType toNumber() const
	{
		NumberType result = 0;
		std::istringstream stream(data);

		if(stream >> result)
			return result;
		else if(data == "yes" || data == "true")
			return 1;

		return 0;
	}

    bool toBool() const
    {
		if(data == "yes" || data == "true" || data == "t" || data == "1") return 1;
		return 0;
	}

    double toDouble() const
    {
        return toNumber<double>();
    }

    int toInt() const
    {
        return toNumber<int>();
    }

	std::string toString() const
	{
		return data;
	}
};
 

/**
* The Config class can be used to load simple key/value pairs from a file.
*
* @note An example of syntax:
*	// An example of a comment
*	username: Bob 
*	gender: male 
*	hair-color: black // inline comments are also allowed
*	level: 42
*
* @note An example of usage:
*	Config config;
*	config.load("myFile.txt");
*	
*	std::string username = config["username"].toString();
*	int level = config["level"].toNumber<int>();
*/

class Config
{
public:

    // create instance of configuration using command-line parameters as priority over config fila loaded later 
	Config(int argc=0, char* argv[]=NULL);

	// Loads key/value pairs from a file
	// Returns whether or not this operation was successful
	bool load(const std::string& filename="");

	// Use the [] operator to access values just like a map container
	// If a key does not exist, will return an empty StringVariant
	const StringVariant& operator[](const std::string& keyName) const;

	// Use the [] operator to get/set values just like a map container
    StringVariant& operator[](const std::string& keyName);

	// extended syntax allows to return a default value 
    const StringVariant& value(const std::string& keyName, const StringVariant& defaultValue="");

    // The actual map used to store key/value pairs
	std::map<std::string, StringVariant> map;

    // config file name if encountered in command line options, overrides that given to load()
    std::string defaultFilename;
};

 // Utility function to trim whitespace off the ends of a string
inline std::string trim(std::string source)
{
	std::string result = source.erase(source.find_last_not_of(" \t") + 1);
	return result.erase(0, result.find_first_not_of(" \t"));
}

Config::Config(int argc, char *argv[])
{
    const std::string comment = "#";
	const std::string delimiter = "=";
    const std::string iniparam = "ini";
    for(int i=1; i<argc; i++)
    {
        std::string arg(argv[i]);
        size_t delimIndex=arg.find(delimiter);
        if(delimIndex != std::string::npos) 
        {
            std::string key = trim(arg.substr(0, delimIndex));
            std::transform(key.begin(), key.end(), key.begin(), ::tolower);
            std::string value = trim(arg.substr(delimIndex + 1));
            if(!key.empty())
                if(key==iniparam)
                    defaultFilename=value;
                else
                    map[key] = value;
        }
    }
}
bool Config::load(const std::string& filename)
{
    const std::string comment = "#";
	const std::string delimiter = "=";
    const std::string iniparam = "ini";
	std::string line;

    std::ifstream file(defaultFilename.empty() ? filename.c_str() : defaultFilename.c_str());

	if(!file.is_open())
		return false;

	while(!file.eof())
	{
		getline(file, line);
		
		// Remove any comments
		size_t commIdx = line.find(comment);
		if(commIdx != std::string::npos)
			line = line.substr(0, commIdx);

		size_t delimIdx = line.find(delimiter);
		if(delimIdx == std::string::npos)
			continue;

		std::string key = trim(line.substr(0, delimIdx));
        std::transform(key.begin(), key.end(), key.begin(), ::tolower);
		std::string value = trim(line.substr(delimIdx + 1));

		if(!key.empty() && map.find(key) == map.end())   // do not overwrite values set from command line
			map[key] = value;
	}

	file.close();

	return true;
}

StringVariant& Config::operator[](const std::string& keyName) 
{ 
    std::string keylower=keyName; 
    std::transform(keylower.begin(), keylower.end(), keylower.begin(), ::tolower);
    return map[keylower]; 
}

const StringVariant& Config::value(const std::string& keyName, const StringVariant& defaultValue)
{
    std::string keylower=keyName; 
    std::transform(keylower.begin(), keylower.end(), keylower.begin(), ::tolower);
    if(map.find(keylower) == map.end())
        return defaultValue;
    else
        return map[keylower]; 
}

const StringVariant& Config::operator[](const std::string& keyName) const 
{
	std::map<std::string, StringVariant>::const_iterator iter = map.find(keyName);
	
	if(iter != map.end())
		return iter->second;

	static StringVariant empty;

	return empty;
}