/** \file   dif1d.cpp
    \brief  provides analytical solution to 1d diffusion problem in a cylindrical geometry with inner sink
    \author EV
    \date   2011-2014
    \version 1.1

This program computes the flux in one-dimensional diffusion problem in a cylindrical geometry, 
in application to studying the loss-cone problem of consumption of stars by a supermassive 
black hole.
The evolution of the distribution function \f$ f(R,t) \f$ in the interval 
\f$ R_{min} <= R <= 1 \f$ is governed by the equation:
\f[
  \frac{ \partial f(R,t) }{ \partial t} = 
  \frac{\partial}{\partial R} \left( D\, R\, \frac{\partial f(R,t)}{\partial R} \right) ,
\f]
with initial conditions \f$ f(R,0) = 1 \f$, subject to the boundary conditions 
\f[
  f(R_{min},t) = \alpha\, R_{min}\, \partial f/\partial R ; \quad
  \frac{\partial}{\partial R} f(1,t) = 0 .
\f]

The solution for the case \f$ \alpha=0 \f$ is given by [Milosavljevic&Merritt 2003] in terms of 
series of Bessel functions (MM03, Eqs.24-28). 
The generalization to arbitrary \f$ \alpha \f$ is done by introducing the effective value 
of \f$ R_0 = R_{min} \exp(-\alpha) \f$ at which the distribution function would be zero. 
The general form of the boundary condition interpolates between empty 
(\f$ \alpha \ll 1 \f$) and full (\f$ \alpha \gg 1 \f$) regimes.

This method is also used to estimate the capture rates in axisymmetric galactic nuclei, 
as described by an effective one-dimensional approximation of [Vasiliev&Merritt 2013].
In this case, the meaning of \f$ R_{min} \f$ is the effective capture boundary \f$ R_{eff} \f$, 
defined in terms of flattening coefficient \f$ R_{sep} \f$ (Eq.72 of VM13).

In addition, the draining of loss region is accounted for with an approximate method described 
in section 6 of [Vasiliev&Merritt 2013]:
initially, the draining flux is equal to the full loss cone capture rate, and declines with time 
so that the total captured mass from the draining region approaches Rdrain. At the same time, 
the flux due to relaxation is gradually scaled up from 0 to the flux in the time-dependent 
solution with initially empty loss region.

Version history:
1.0 - first release, for the paper [Vasiliev&Merritt 2013, ApJ]
1.1 - corrected computation of draining flux, for the paper [Vasiliev 2014, CQG]

*/

#include <cmath>
#include <vector>
#include <iostream>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_bessel.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_integration.h>
#include "config.h"
double gsl_pow_2(const double x) { return x*x; }

// global variables that are initialized in the main function and are used in other fncs
// parameters that are initialized from the command line
double D=1;               // diffusion coefficient D. Note that in MM03 a quantity mu is used (mu=D/4)
double alpha=0;           // coefficient in the boundary condition (<<1 -- empty, >>1 -- full loss cone regimes)
double Rmin=0;            // capture boundary (squared normalized angular momentum of the capture orbit)
double Rdrain=0;          // angular momentum of draining region (= max.ang.mom.of centrophilic orbits)
double Fdrain=1;          // initial amplitude of distribution function in the draining region (0 - no draining, 1 - initially isotropic d.f. everywhere)
double FluxFullLC=0;      // amplitude of the flux of the full loss cone (limits the maximum value of the capture rate)
double maxtime=0;         // total time interval for the solution
double tstep=0;           // timestep at which the output is displayed
// other variables computed from the above parameters
double norm=1;            // normalization of the function (=1-Rmin)
double lc0=0.01;          // value of j_{lc,0}=sqrt(R_0) at which the d.f. equals to zero (R0=Rmin*exp(-alpha))
double Tdrain=0;          // draining time of the draining region (=Rdrain/FluxFullLC)
const int MMAX=100;       // number of Bessel functions in the expansion of the initial d.f. (m=1..MMAX)
std::vector<double> beta; // roots of Bessel functions
std::vector<double> coef; // coefficients of the expansion

// a replacement for the default error handler that just displays error message and doesn't crash program
static void gsl_verbose_error_handler (const char *reason, const char *file, int line, int gsl_errno)
{
    if(gsl_errno == GSL_ETOL) return;  // do nothing
    std::cerr << "GSL error "<<gsl_errno<<" in "<<file<<" at line "<<line<<": "<<reason<<"\n";
    std::cerr << "Rmin="<<Rmin<<", Rdrain="<<Rdrain<<", alpha="<<alpha<<",D="<<D<<"\n";
    exit(1);
}

// function A(x,y) defined by Eq.25 in MM03
double fncA(double x, void* params)
{
    double y=*((double*)params);
    return gsl_sf_bessel_J0(x*y)*gsl_sf_bessel_Y1(x) - gsl_sf_bessel_Y0(x*y)*gsl_sf_bessel_J1(x);
}
// the product A(x,y)*y that enters the expression for the expansion coefficients (Eq.24 in MM03)
double fncAy(double y, void* params)
{
    double x=*((double*)params);
    return (gsl_sf_bessel_J0(x*y)*gsl_sf_bessel_Y1(x) - gsl_sf_bessel_Y0(x*y)*gsl_sf_bessel_J1(x))*y;
}
// function B(x,y) defined by Eq.28 in MM03
double fncB(double x, void* params)
{
    double y=*((double*)params);
    return gsl_sf_bessel_J1(x*y)*gsl_sf_bessel_Y1(x) - gsl_sf_bessel_Y1(x*y)*gsl_sf_bessel_J1(x);
}
// the solution N(j,t) as given by Eq.24 of MM03. 
// Note that we do not actually use this expression in what follows, 
// but rather integrate the flux numerically
double fncN(double j, double t)
{
    double result=0;
    for(int m=0; m<MMAX; m++)
        result += coef[m] * fncA(beta[m], &j) * exp(-gsl_pow_2(beta[m])*D/4*t);
    return result;
}
// the expression for the flux through the loss cone boundary F(t) as given by Eq.27 of MM03
double FluxNoDrain(double t)
{
    if(alpha>500)
        return FluxFullLC;  // to avoid numerical overflow, we use a constant full loss cone flux for large alpha
    double result=0;
    for(int m=0; m<MMAX; m++)
        result += beta[m]*coef[m] * fncB(beta[m], &lc0) * exp(-gsl_pow_2(beta[m])*D/4*t);
    result*=-norm*D*lc0/2;
    return result;
}
// the flux (capture rate) from draining of centrophilic orbits, 
// as given by Eq.74b of VM13
double FluxDrain(double t)
{
    double z=2*t/Tdrain;  // 2 * Eq.75 of VM13
    if(z<1e-6) return Fdrain*FluxFullLC*(1-z*2/3);  // asymptotic expression for small t that avoids roundoff errors
    return Tdrain>0? Fdrain*FluxFullLC*(1-(1+z)*exp(-z))*2/z/z : 0;
}
// time integral of the above (Eq.74a of VM13)
double FluxDrainTotal(double t)
{
    double z=2*t/Tdrain;
    if(z<1e-6) return z/2;
    return Tdrain>0? (1- (1-exp(-z))/z) : 1;
}
// here comes the kludge: 
// we want an interpolating formula that is equivalent to the draining rate at small t 
// and to the time-dependent flux of the diffusion problem at large t.
// We multiply the latter quantity by the fraction of captured mass from the centrophilic orbits, 
// which varies from 0 at t=0 to 1 at t>>T_drain, and add the draining flux.
// This expression is not well motivated but gives a correct asymptotical behaviour at both small and large t.
double FluxWithDrain(double t, void *params)
{
    return std::min(FluxFullLC, FluxNoDrain(t)*FluxDrainTotal(t) + FluxDrain(t));
}

int main(int argc, char* argv[])
{
    gsl_set_error_handler(&gsl_verbose_error_handler);
    Config params(argc, argv);
    alpha= params.value("alpha",0).toDouble();      // coef-t in the boundary condition
    Rmin = params.value("Rmin",-1).toDouble();      // (effective) capture boundary
    Rdrain=params.value("Rdrain",Rmin).toDouble();  // draining region
    D=params.value("dc",1).toDouble();              // diffusion coefficient
    FluxFullLC=params.value("MaxFlux",D/alpha).toDouble();  // full loss cone flux (limiting the initial flux in the diffusion problem)
    if(FluxFullLC==0) FluxFullLC=gsl_posinf();      // unbound
    Fdrain=params.value("Fdrain",1).toDouble();     // value of d.f. in the drain region (1 means it has initially full loss cone flux)
    Tdrain=Rdrain/FluxFullLC;                       // draining time of the drain region (to decrease d.f. by a factor of ~2)
    if(Rmin<=0 || Rmin>=1) { std::cerr << "Error, Rmin should be >0 and <1\n"; return 0; }
    else lc0=sqrt(Rmin*exp(-alpha));
    if(Fdrain<0 || Fdrain>1) { std::cerr << "Error, Fdrain should be >=0 and <=1\n"; return 0; }
    if(Rdrain<0 || Rdrain>1) { std::cerr << "Error, Rdrain should be >=0 and <=1\n"; return 0; }
    norm =1-Rmin;
    maxtime=params.value("tmax",1).toDouble();  // total time of the solution
    tstep=params.value("tstep",maxtime/20).toDouble();  // output interval 
    maxtime*=1.0000001;
    
    gsl_function F;
    gsl_root_fsolver* solv = gsl_root_fsolver_alloc(gsl_root_fsolver_brent); 

    // to avoid numerical overflow, we use the series solution only if alpha is not too large;
    // otherwise use the full loss cone flux
    if(alpha<500) {
        // find roots beta_m given implicitly by Eq.26 of MM03
        F.function=&fncA;
        F.params=&lc0;
        double betalow=0.125, betaupp=3.0, betaroot;
        while(fncA(betalow,&lc0)>0) betalow/=2;
        while(fncA(betaupp,&lc0)<0) betaupp+=0.1;
        for(int m=0; m<MMAX; m++)
        {
            int status=0, iter=0;
            gsl_root_fsolver_set (solv, &F, betalow, betaupp);
            do{
                iter++;
                status  = gsl_root_fsolver_iterate (solv);
                betaroot= gsl_root_fsolver_root (solv);
                betalow = gsl_root_fsolver_x_lower (solv);
                betaupp = gsl_root_fsolver_x_upper (solv);
                status = gsl_root_test_interval (betalow, betaupp, 0, 1e-8);
            }
            while(status == GSL_CONTINUE && iter < 100); 
            beta.push_back(betaroot);
            betalow=betaroot+0.1;
            double sign=fncA(betalow, &lc0);
            betaupp=betaroot+4;
            while(fncA(betaupp, &lc0)*sign > 0) betaupp+=0.1;
        }
        // find the coefficients of expansion of the initial d.f. N(j,t=0) over the Bessel basis.
        // each coefficient is given by the integral \int_{j_{lc0}}^1 j' A(\beta_m,j') N(j',t=0) dj'
        // multiplied by some combination of Bessel functions -- in other words, all terms in 
        // Eq.24 of MM03 that do not depend on j and t.
        F.function=&fncAy;
        gsl_integration_workspace* ws = gsl_integration_workspace_alloc(1000);
        for(int m=0; m<MMAX; m++)
        {
            F.params=&(beta[m]);
            double result, error;
            gsl_integration_qag(&F, lc0, 1, 0, 1e-6, 1000, GSL_INTEG_GAUSS21, ws, &result, &error);
            double j0=gsl_pow_2(gsl_sf_bessel_J0(beta[m]*lc0));
            result *= gsl_pow_2(beta[m])*j0 / (j0 - gsl_pow_2(gsl_sf_bessel_J1(beta[m]))) * M_PI*M_PI/2;
            coef.push_back(result);
        }
    }
    // output instantaneous capture rate and total captured mass as functions of time
    double FluxTotal=0;
    F.function=&FluxWithDrain;
    for(double t=tstep; t<=maxtime; t+=tstep) {
        double result, error; size_t neval;
        // total captured mass during this interval of time is obtained by integrating the flux.
        // since we use a weighted sum of draining and relaxation fluxes, we can't use 
        // analytic expressions for time integral of either of them, 
        // and perform the integration numerically.
        gsl_integration_qng(&F, t-tstep, t, 0, 1e-6, &result, &error, &neval);
        FluxTotal+=result;
        std::cout << t << "\t" << FluxWithDrain(t, NULL) << "\t" << FluxTotal << "\t" << FluxDrain(t) << "\t" << (Fdrain*Rdrain*FluxDrainTotal(t)) <<"\n";
    }
    return 0;
}

