#include "massmodel.h"
#include <cmath>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include "utils.h"
#ifdef DEBUGPRINT
#include <iostream>
#endif

namespace smile{

//----------------------------------------------------------------------------//
inline double pow_2(const double x) { return x*x; }

// interpolated mass inside radius
double CMassModel::mass(double r) const
{
    if(r==0) return masscenter;
    if(r>=rout)  // special treatment for extrapolation beyond the last grid point
        return totalmass - coefmrout*pow(r, 3-gammaout);
    double mu = exp(gsl_spline_evalx(spl_loginmass, log(r), acc_loginmass));
    return (mu*totalmass + masscenter)/(mu+1);
}

// interpolated radius as a function of enclosed mass
double CMassModel::rad(double m) const
{
    if(m<=masscenter) return 0;
    if(m>=totalmass) return gsl_posinf();
    return exp(gsl_spline_evalx(spl_logrmass, log( (m-masscenter)/(totalmass-m) ), acc_logrmass));
}

// interpolated density at radius
double CMassModel::dens(double r) const
{
    if(densconst && r<rin) return rho0*(1-coefdrhodr*pow(r, alpha));
    if(r>rout) return (gammaout-3)/(4*M_PI)*coefmrout*pow(r, -gammaout);
    double inm=mass(r);
    return gsl_spline_eval_derivx(spl_loginmass, log(r), acc_loginmass) * (inm-masscenter) * (totalmass-inm) / (totalmass-masscenter) / (4*M_PI*pow(r, 3.0));
}

// interpolated potential at radius
double CMassModel::pot(double r) const
{
    if(r==0) return !masscenter ? potentialcenter : gsl_neginf();
    if(r<=rin)
    {
        double phi1=potentialin+masscenter/rin;
        return potentialcenter + (phi1-potentialcenter)*pow(r/rin, 2-gammain) - masscenter/r;
    }
    double phi = -exp(gsl_spline_evalx(spl_logpotential, log(r), acc_logpotential));
    if(!masscenter) return 1/(1/phi+1/potentialcenter); else return phi;
}

// interpolated distribution function of energy, given by pre-computed spline via 2nd derivative of rho(Phi)
double CMassModel::distrf(double phi) const
{
    double logpot = !masscenter ? log(-1/(1/phi-1/potentialcenter)) : log(-phi);
    return exp(gsl_spline_evalx(spl_logdistrf, logpot, acc_logdistrf));
}

// interpolated distribution function of energy, given by differentiation of pre-computed spline via 1st derivative of rho(Phi)
double CMassModel::distrf1(double phi) const
{
    if(phi>=0) return 0;
    double logpot = !masscenter ? log(-1/(1/phi-1/potentialcenter)) : log(-phi);
    double distrfint = -exp(gsl_spline_evalx(spl_logintdistrf, logpot, acc_logintdistrf));
    return std::max<double>(0, 
        gsl_spline_eval_derivx(spl_logintdistrf, logpot, acc_logintdistrf) * distrfint / phi / (!masscenter ? 1-phi/potentialcenter : 1) );
}

// interpolated differential energy distribution
double CMassModel::ded(double phi) const
{
    if(phi>=0) return 0;
    double logpot = !masscenter ? log(-1/(1/phi-1/potentialcenter)) : log(-phi);
    return exp(gsl_spline_evalx(spl_logded, logpot, acc_logded));
}

// interpolated radial period at energy
double CMassModel::trad(double phi) const
{
    double logpot = !masscenter ? log(-1/(1/phi-1/potentialcenter)) : log(-phi);
    return exp(gsl_spline_evalx(spl_logtrad, logpot, acc_logtrad));
}

// interpolated radius of circular orbit at energy
double CMassModel::rcirc(double phi) const
{
    double logpot = !masscenter ? log(-1/(1/phi-1/potentialcenter)) : log(-phi);
    return exp(gsl_spline_evalx(spl_logrcirc, logpot, acc_logrcirc));
}

// interpolated angular momentum of circular orbit at energy
double CMassModel::lcirc(double phi) const
{
    double rc=rcirc(phi);
    return sqrt(2*(phi-pot(rc)))*rc;
}

/* --- helper functions for integration and root-finding --- */

/// \cond INTERNAL_DOCS
// helper structure keeping params for GSL call-back functions
struct CHelperParam{
    const CMassModel* model;
    double E;
    double M;
    double Phi;
    double L2;
    double kprime_over_k;
    double n;
    double p2,p3,r1,r2,r3;  // for constant-density core estimator
};
/// \endcond 

// integrand for potential
double int_pot(double r, void* params)
{
    if(r==0) return 0;
    return (((CHelperParam*)params)->model->mass(r) - ((CHelperParam*)params)->model->mass(0))/r/r;
}
// integrand for Eddington inversion formula, using first or second derivative variant
double int_distr1(double Phi, void* params)
{
    return ((CHelperParam*)params)->model->drhodphi(Phi)/sqrt(Phi-((CHelperParam*)params)->E);
}
double int_distr2(double Phi, void* params)
{
    return ((CHelperParam*)params)->model->d2rhodphi2(Phi)/sqrt(Phi-((CHelperParam*)params)->E);
}
// integrand for finding radial period of purely radial orbit
double int_trad(double r, void* params)
{
    //return 1/sqrt(2*( ((CHelperParam*)params)->E - ((CHelperParam*)params)->model->pot(r)));
    double pot=((CHelperParam*)params)->model->pot(r);
    return 1/sqrt(2*( ((CHelperParam*)params)->E - pot));;
}
// integrand for differential energy distribution
double int_ded(double r, void* params)
{
    if(r==0) return 0;
    return r*r*sqrt(std::max<double>(2*( ((CHelperParam*)params)->E - ((CHelperParam*)params)->model->pot(r)),0));
}
// integrand for f*g
double int_fg(double E, void* params)
{
    return ((CHelperParam*)params)->model->ded(E) * ((CHelperParam*)params)->model->distrf(E);
}
// integrand for <v^n(r)>
double int_sigma(double v, void* params)
{
    return ((CHelperParam*)params)->model->distrf(((CHelperParam*)params)->Phi+v*v/2) *4*M_PI*pow(v, 2+((CHelperParam*)params)->n);
}
// integrand for projected density
double int_densproj(double z, void* params)
{
    return ((CHelperParam*)params)->model->dens(sqrt(pow_2(z)+pow_2(((CHelperParam*)params)->r1)));
}
// integrand for line-of-sight velocity dispersion
double int_losvd(double z, void* params)
{
    double r=sqrt(pow_2(z)+pow_2(((CHelperParam*)params)->r1));
    return ((CHelperParam*)params)->model->dens(r)*((CHelperParam*)params)->model->mass(r)*z*z/pow(r,3);
}
// helper root-finder for radius as a function of potential
double find_pot(double r, void* params)
{
    return ((CHelperParam*)params)->model->pot(r)-((CHelperParam*)params)->Phi;
}
// helper root-finder for a constant-density core
double find_alpha(double a, void* params)
{
    return (1-((CHelperParam*)params)->p2)*(pow(((CHelperParam*)params)->r3,a)-((CHelperParam*)params)->p3*pow(((CHelperParam*)params)->r1,a)) 
         - (1-((CHelperParam*)params)->p3)*(pow(((CHelperParam*)params)->r2,a)-((CHelperParam*)params)->p2*pow(((CHelperParam*)params)->r1,a));
}
// helper root-finder for circular radius at energy
double find_rcirc(double r, void* params)
{
    return ((CHelperParam*)params)->model->pot(r) + ((CHelperParam*)params)->model->mass(r)/2/r - ((CHelperParam*)params)->E;
}
// helper root-finder for radius at mass
double find_radius_by_mass(double r, void* params)
{
    return ((CHelperParam*)params)->model->mass(r) - ((CHelperParam*)params)->M;
}
// helper root-finder for maximum of f()*v^2
double find_max_fv(double v, void* params)
{
    return -((CHelperParam*)params)->model->distrf(v*v/2 - ((CHelperParam*)params)->Phi)*v*v;
}
// helper function for finding outer density slope
double find_gamma_out(double y, void* params)
{
    return( pow(((CHelperParam*)params)->r2, 3-y) - pow(((CHelperParam*)params)->r1, 3-y))
        / ( pow(((CHelperParam*)params)->r3, 3-y) - pow(((CHelperParam*)params)->r1, 3-y))
        - ((CHelperParam*)params)->M;
}

// rmax: r(Phi)
double CMassModel::rmax(double phi) const
{
    CHelperParam params;
    params.Phi=phi;
    params.model=this;
    gsl_function F;
    F.function=&find_pot;
    F.params=&params;
    return findRoot(&F, -masscenter/phi, -totalmass/phi);
}

// interpolated d(rho)/d(phi) at given phi
double CMassModel::drhodphi(double phi) const
{
    if(phi<potentialin) {
        if(densconst) return coefdrhodphi*pow(phi-potentialcenter, alpha/2-1);
        double rad=0;  // find radius corresponding to phi
        double massrin=mass(rin)-masscenter;
        if(masscenter==0) 
            rad=rin*pow( (2-gammain)*(phi-potentialcenter)*rin/massrin, 1./(2-gammain));
        else { // no analytical expression -- need to solve numerically
            CHelperParam params;
            params.Phi=phi;
            params.model=this;
            gsl_function F;
            F.function=&find_pot;
            F.params=&params;
            rad=findRoot(&F, -masscenter/phi, rin);
        }
        double massr=massrin*pow(rad/rin, 3-gammain);
        double dphi1dr1=(masscenter+massr)/(rad*rad);
        double drho1dr1=-gammain*(3-gammain)/(4*M_PI)*massr/pow(rad,4);
        return drho1dr1/dphi1dr1;
    }
    double logpot = !masscenter ? log(-1/(1/phi-1/potentialcenter)) : log(-phi);
    double rho = exp(gsl_spline_evalx(spl_logdenspot, logpot, acc_logdenspot));
    return gsl_spline_eval_derivx(spl_logdenspot, logpot, acc_logdenspot) * rho / phi / (!masscenter ? 1-phi/potentialcenter : 1);
}

// interpolated d^2(rho)/d(phi)^2 at given phi
double CMassModel::d2rhodphi2(double phi) const
{
    if(phi<potentialin) {
        if(densconst) return coefdrhodphi*(alpha/2-1)*pow(phi-potentialcenter, alpha/2-2);
        double rad=0;  // find radius corresponding to phi
        double massrin=mass(rin)-masscenter;
        if(masscenter==0) 
            rad=rin*pow( (2-gammain)*(phi-potentialcenter)*rin/massrin, 1./(2-gammain));
        else { // no analytical expression -- need to solve numerically
            CHelperParam params;
            params.Phi=phi;
            params.model=this;
            gsl_function F;
            F.function=&find_pot;
            F.params=&params;
            rad=findRoot(&F, -masscenter/phi, rin);
        }
        double massr=massrin*pow(rad/rin, 3-gammain);
        double dphi1dr1=(masscenter+massr)/(rad*rad);
        double dphi2dr2=(-2*masscenter+massr*(1-gammain))/pow(rad,3);
        double drho1dr1=-gammain*(3-gammain)/(4*M_PI)*massr/pow(rad,4);
        double drho2dr2=-(gammain+1)*drho1dr1/rad;
        return (drho2dr2*dphi1dr1-drho1dr1*dphi2dr2)/pow(dphi1dr1,3);
    }
    double logpot = !masscenter ? log(-1/(1/phi-1/potentialcenter)) : log(-phi);
    double rho = exp(gsl_spline_evalx(spl_logdenspot, logpot, acc_logdenspot));
    double drho1dphi1 = gsl_spline_eval_derivx(spl_logdenspot, logpot, acc_logdenspot);
    double drho2dphi2 = gsl_spline_eval_deriv2x(spl_logdenspot, logpot, acc_logdenspot);
    return rho / pow_2(phi * (!masscenter ? 1-phi/potentialcenter : 1)) *
        (drho2dphi2 + pow_2(drho1dphi1) - drho1dphi1*( !masscenter ? 1-2*phi/potentialcenter : 1) );
}

// direct evaluation of distribution function via Eddington inversion
double CMassModel::distrf2(double E) const
{
    gsl_function F;
    CHelperParam params;
    params.model=this;
    F.function=&int_distr2;
    F.params=&params;
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc(1000);
    params.E=E;
    double result, error;
    gsl_integration_qags (&F, 0, params.E, 0, 1e-10, 1000, ws, &result, &error); 
    gsl_integration_workspace_free(ws);
    return -(sqrt(0.125)/M_PI/M_PI) * result;
}

// compute velocity dispersion at given radius
double CMassModel::veldisp(double rad) const
{
    gsl_integration_workspace* ws = gsl_integration_workspace_alloc(1000);
    CHelperParam params;
    params.Phi=pot(rad);
    params.model=this;
    gsl_function F;
    F.function=&int_sigma;
    F.params=&params;
    double result, result0, error, vmax=sqrt(-2*params.Phi);
    params.n=2;
    gsl_integration_qags(&F, 0, vmax, 0, 1e-7, 1000, ws, &result, &error);
    params.n=0;
    gsl_integration_qags(&F, 0, vmax, 0, 1e-7, 1000, ws, &result0, &error);
    error=dens(rad);
    gsl_integration_workspace_free(ws);
    return sqrt(result/result0/3.0 /*dens(rad)*/);
}

// compute projected density at given radius
double CMassModel::densproj(double rad) const
{
    gsl_integration_workspace* ws = gsl_integration_workspace_alloc(1000);
    CHelperParam params;
    params.r1=rad;
    params.model=this;
    gsl_function F;
    F.function=&int_densproj;
    F.params=&params;
    double result, error;
    gsl_integration_qagiu(&F, 0, 0, 1e-7, 1000, ws, &result, &error);
    gsl_integration_workspace_free(ws);
    return 2*result;
}

// compute line-of-sight velocity dispersion at a given projected radius
double CMassModel::losvd(double rad) const
{
    gsl_integration_workspace* ws = gsl_integration_workspace_alloc(1000);
    CHelperParam params;
    params.r1=rad;
    params.model=this;
    gsl_function F;
    F.function=&int_losvd;
    F.params=&params;
    double result, error;
    gsl_integration_qagiu(&F, 0, 0, 1e-7, 1000, ws, &result, &error);
    gsl_integration_workspace_free(ws);
    return sqrt(2*result/densproj(rad));
}

// compute mass of particles having energy <E
double CMassModel::masse(double E) const
{
    gsl_integration_workspace* ws = gsl_integration_workspace_alloc(1000);
    double elow=!masscenter?potentialcenter:potentialin*100;
    double result, error;
    gsl_function F;
    CHelperParam params;
    F.function=&int_fg;
    F.params=&params;
    params.model=this;
    gsl_integration_qags(&F, elow, E, 0, 1e-7, 1000, ws, &result, &error);
    gsl_integration_workspace_free(ws);
    return result;
}

// generate particle from f(E) at given r
double CMassModel::samplev(double radius) const
{
    if(!gsl_finite(radius)) return 0;  // infinity
    gsl_function F;
    CHelperParam params;
    F.function=&find_max_fv;
    F.params=&params;
    params.model=this;
    params.Phi=-pot(radius);
    // determine maximum of v^2*f(Psi-v^2/2)
    double vesc=sqrt(2*params.Phi);
    double vroot = findMin(&F, vesc*1e-8, vesc*(1-1e-8), vesc/2, 1e-3);
    double fmax = -find_max_fv(vroot, &params) * 1.1;
    double v, f1;
    do{
        v=vesc*rand()/RAND_MAX;//gsl_rng_uniform(randn);
        f1=fmax*rand()/RAND_MAX;//gsl_rng_uniform(randn);
    } while(f1>distrf(-params.Phi+v*v/2)*v*v);
    return v;
}

/* --- Calculation of diffusion coefficients --- */

// integrand for I0
double int_I0(double Ef, void* params)
{
    return ((CHelperParam*)params)->model->distrf(-Ef);
}
// integrand for In
double int_In(double Ef, void* params)
{
    return ((CHelperParam*)params)->model->distrf(-Ef) * pow( ((CHelperParam*)params)->Phi-Ef, ((CHelperParam*)params)->n);
}
// root-finder for v_r^2=0
double find_vr2(double r, void* params)
{
    return 2*(-((CHelperParam*)params)->model->pot(r) - ((CHelperParam*)params)->E) - ((CHelperParam*)params)->L2/(r*r);
}
// coefficients appearing in delta v_par, delta v_per
double coef_I0(void* params)
{
    double result, error; size_t niter;
    gsl_function F;
    F.function=&int_I0;
    F.params=params;
    gsl_integration_qng(&F, 0, ((CHelperParam*)params)->E, 0, 1e-4, &result, &error, &niter);
    return result*16*M_PI*M_PI;
}
double coef_In(void* params, double n)
{
    double result, error; size_t niter;
    gsl_function F;
    F.function=&int_In;
    F.params=params;
    ((CHelperParam*)params)->n=n;
    gsl_integration_qng(&F, ((CHelperParam*)params)->E, ((CHelperParam*)params)->Phi, 0, 1e-4, &result, &error, &niter);
    return result*16*M_PI*M_PI / pow(((CHelperParam*)params)->Phi-((CHelperParam*)params)->E, n);
}
// local diffusion coefs (E > 0 is binding energy, R is squared relative ang.mom.)
double Dif1(double r, void* params)
{
    double vr2=find_vr2(r, params);
    if(vr2<=0) return 0;
    return 1 / sqrt(vr2);
}
double DifEElocal(double r, void* params)
{
    double Phi=-((CHelperParam*)params)->model->pot(r);
    double E=((CHelperParam*)params)->E;
    double vr2=find_vr2(r, params);
    if(vr2<=0) return 0;
    ((CHelperParam*)params)->Phi=Phi; 
    return 2*(Phi-E) * 2.0/3 * (coef_I0(params) + coef_In(params, 1.5)) / sqrt(vr2);
}
double DifRRlocal(double r, void* params)
{
    double Phi=-((CHelperParam*)params)->model->pot(r);
    double E=((CHelperParam*)params)->E;
    //double L2=((CHelperParam*)params)->L2;
    double vr2=find_vr2(r, params);
    if(vr2<=0) return 0;
    ((CHelperParam*)params)->Phi=Phi; 
    double v2=2*(Phi-E);
    double mulv2par=pow_2(2+((CHelperParam*)params)->kprime_over_k*v2);
    double mulv2per=2*vr2/(v2-vr2);
    return 2.0/3 * (coef_I0(params)*(mulv2par+2*mulv2per) + 3*coef_In(params, 0.5)*mulv2per + coef_In(params, 1.5)*(mulv2par-mulv2per)) / v2 / sqrt(vr2);
}
double DifLLlocal(double r, void* params)
{
    double Phi=-((CHelperParam*)params)->model->pot(r);
    double E=((CHelperParam*)params)->E;
    //double L2=((CHelperParam*)params)->L2;
    double vr2=find_vr2(r, params);
    if(vr2<=0) return 0;
    ((CHelperParam*)params)->Phi=Phi; 
    double v2=2*(Phi-E);
    double mulv2par=1;
    double mulv2per=0.5*vr2/(v2-vr2);
    return 2.0/3 * (coef_I0(params)*(mulv2par+2*mulv2per) + 3*coef_In(params, 0.5)*mulv2per + coef_In(params, 1.5)*(mulv2par-mulv2per)) / v2 / sqrt(vr2);
}
// L-averaged diffusion coefs (E > 0 is binding energy)
double Dif1E(double r, void* params)
{
    double v2=2*(-((CHelperParam*)params)->model->pot(r) - ((CHelperParam*)params)->E);
    if(v2<=0) return 0;
    return sqrt(v2)*r*r;
}
double DifEElocE(double r, void* params)
{
    double Phi=-((CHelperParam*)params)->model->pot(r);
    double E=((CHelperParam*)params)->E;
    double v2=2*(Phi-E);
    if(v2<=0) return 0;
    ((CHelperParam*)params)->Phi=Phi; 
    return pow(v2,1.5)*r*r * 2.0/3 * (coef_I0(params) + coef_In(params, 1.5));
}
double DifLLlocE(double r, void* params)
{
    double Phi=-((CHelperParam*)params)->model->pot(r);
    double E=((CHelperParam*)params)->E;
    double v2=2*(Phi-E);
    if(v2<=0) return 0;
    ((CHelperParam*)params)->Phi=Phi; 
    return sqrt(v2)*r*r * r*r/3 * (2*coef_I0(params) + coef_In(params, 0.5) + coef_In(params, 1.5));
}
double DifELlocE(double r, void* params)
{
    double Phi=-((CHelperParam*)params)->model->pot(r);
    double E=((CHelperParam*)params)->E;
    double v2=2*(Phi-E);
    if(v2<=0) return 0;
    ((CHelperParam*)params)->Phi=Phi; 
    return v2*pow(r,3) *M_PI/6 *(coef_I0(params) + coef_In(params, 1.5));
}
double DifRRlocE(double r, void* params)
{
    double Phi=-((CHelperParam*)params)->model->pot(r);
    double E=((CHelperParam*)params)->E;
    double v2=2*(Phi-E);
    if(v2<=0) return 0;
    ((CHelperParam*)params)->Phi=Phi; 
    double Q2=pow_2(1+((CHelperParam*)params)->kprime_over_k*v2/2);
    return pow(v2,1.5)*pow(r,6) * (coef_I0(params)*(8*Q2+2) + coef_In(params, 0.5)*3 + coef_In(params, 1.5)*(8*Q2-1))*8.0/45; 
}
double DifLLlocEfix(double r, void* params)
{
    double Phi=-((CHelperParam*)params)->model->pot(r);
    double E=((CHelperParam*)params)->E;
    double v2=2*(Phi-E);
    if(v2<=0) return 0;
    return pow(r,4.0)*((CHelperParam*)params)->model->dens(r);
}

// orbit-averaged dif.coefs at given E, R
void CMassModel::DifCoefAvER(double E, double R, double *DEE, double *DRR, double *DLL) const
{
    E*=-1;  // use positive(binding) energy throughout this routine
    gsl_integration_workspace* ws_difcoefs = gsl_integration_workspace_alloc(1000);
    // find orbit extent (r_min, r_max)
    double rc=rcirc(-E), rmin, rmax;
    gsl_function F;
    CHelperParam params;
    F.function=&find_vr2;
    F.params=&params;
    params.model=this;
    params.E=E;
    params.L2=rc*rc* 2*(-pot(rc)-E) * R;
    double rc1=rcirc(-E*0.9999);
    params.kprime_over_k = (1 - rc1*rc1* 2*(-pot(rc1)-E*0.9999) / (rc*rc* 2*(-pot(rc)-E)))/(0.0001*E);
    rmin = findRoot(&F, rc*R/100, rc);
    rmax = findRoot(&F, rc, rc*4);
    // now integrate local dif.coefs from r_min to r_max
    F.function=&Dif1;
    double result, error;
    gsl_integration_qags (&F, rmin, rmax, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    double trad = result;
    if(DEE){
    F.function=&DifEElocal;
    gsl_integration_qags (&F, rmin, rmax, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    *DEE = result / trad;
    }
    if(DRR){
    F.function=&DifRRlocal;
    gsl_integration_qags (&F, rmin, rmax, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    *DRR = result*R*R / trad;
    }
    if(DLL){
    F.function=&DifLLlocal;
    gsl_integration_qags (&F, rmin, rmax, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    *DLL = result*params.L2 / trad;
    }
    gsl_integration_workspace_free(ws_difcoefs);
}

// fully-averaged (ergodic) dif.coefs (depend on E only)
void CMassModel::DifCoefAvE(double E, double *DEE, double *DRR, double *DLL, double *DEL) const
{
    E*=-1;  // use positive(binding) energy throughout this routine
    gsl_integration_workspace* ws_difcoefs = gsl_integration_workspace_alloc(1000);
    // find orbit extent (r_min, r_max)
    gsl_function F;
    CHelperParam params;
    F.params=&params;
    params.model=this;
    params.E=E;
    double rmx=rmax(-E), rc=rcirc(-E), rc1=rcirc(-E*0.9999), Lc2=rc*rc* 2*(-pot(rc)-E);
    params.kprime_over_k = (1 - rc1*rc1* 2*(-pot(rc1)-E*0.9999) / Lc2)/(0.0001*E);
    // now integrate local dif.coefs from r_min to r_max
    F.function=&Dif1E;
    double result, error;
    gsl_integration_qags (&F, 0, rmx, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    double phasevol = result;
    if(DEE){
    F.function=&DifEElocE;
    gsl_integration_qags (&F, 0, rmx, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    *DEE = result / phasevol;
    }
    if(DRR){
    F.function=&DifRRlocE;
    gsl_integration_qags (&F, 0, rmx, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    *DRR = result / phasevol / pow_2(Lc2);
    }
    if(DLL){
    F.function=&DifLLlocE;
    gsl_integration_qags (&F, 0, rmx, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    *DLL = result / phasevol;
    }
    if(DEL){
    F.function=&DifLLlocEfix;//&DifELlocE;
    gsl_integration_qags (&F, 0, rmx, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
    *DEL = result / phasevol;
    }
    gsl_integration_workspace_free(ws_difcoefs);
}

// lim(DRR/R), R->0
double CMassModel::DRRrad(double E) const
{
    double DRR;
    DifCoefAvER(E, 0.001, NULL, &DRR, NULL);
    return DRR/0.001 / 2;  // D_RR = 1/2 * <Delta R^2>
    // it's not the most fancy way of computing it, I know...
}

// constructor:
// create model from array of { r, m(r) } pairs, initialize all relevant splines
CMassModel::CMassModel(const std::vector<double> &initradii, const std::vector<double> &initmass)
{
    spl_loginmass=NULL;
    spl_logrmass=NULL;
    spl_logpotential=NULL;
    spl_logdenspot=NULL;
    spl_logdistrf=NULL;
    spl_logintdistrf=NULL;
    spl_logded=NULL;
    spl_logtrad=NULL;
    spl_logrcirc=NULL;
    acc_loginmass=NULL;
    acc_logrmass=NULL;
    acc_logpotential=NULL;
    acc_logdenspot=NULL;
    acc_logintdistrf=NULL;
    acc_logdistrf=NULL;
    acc_logded=NULL;
    acc_logtrad=NULL;
    acc_logrcirc=NULL;
    errorcode=MS_OK;
    if(initradii.size()!=initmass.size())
    {
        errorcode=MS_ARRAY_NOT_EQ;  // input arrays not equal
        return;
    }
    for(size_t i=1; i<initradii.size(); i++)
    {
        if(initradii[i]<=initradii[i-1]){
            errorcode=MS_RADII_NOT_ASC; // "Radii not in ascending order!\n";
            return;
        }
        if(initmass[i]<=initmass[i-1]){
            errorcode=MS_MASS_NOT_ASC; // "Mass not increasing with radius!\n";
            return;
        }
    }
    int Npoints=static_cast<int>(initradii.size()), Nskip=0;
    if(initradii.front()<=0)
    {
        masscenter=initmass.front();
        Npoints--;
        Nskip=1;
    }
    else
        masscenter=0;
    if(Npoints<5) {
        errorcode=MS_ARRAY_SHORT;  // Too few points
        return;
    }
    rout=initradii.back();
    rin=initradii[Nskip];
    gsl_function F;
    CHelperParam params;
    params.model=this;
    F.params=&params;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc(1000);

    std::vector<double> logradii, loginmass, logpotential, logdens, logintdistrf, logded, logtrad, logrcirc;
    logradii.resize(Npoints,0);
    loginmass.resize(Npoints,0);
    logintdistrf.resize(Npoints,0);
    logded.resize(Npoints,0);
    logdens.resize(Npoints,0);
    logpotential.resize(Npoints,0);
    logtrad.resize(Npoints,0);
    logrcirc.resize(Npoints,0);

    // estimate outermost slope  (needed for accurate extrapolation beyond last grid point)
    totalmass=0;
    params.r1=initradii[Npoints+Nskip-1];
    params.r2=initradii[Npoints+Nskip-2];
    params.r3=initradii[Npoints+Nskip-3];
    params.M = (initmass[Npoints+Nskip-1] - initmass[Npoints+Nskip-2])/(initmass[Npoints+Nskip-1] - initmass[Npoints+Nskip-3]);
    F.function=&find_gamma_out;
    double Ylow=3.01, Yupp=10;
    gammaout = findRoot(&F, Ylow, Yupp, 5.0);
    totalmass = 
        (initmass[Npoints+Nskip-1]*pow(initradii[Npoints+Nskip-1], gammaout-3) -
         initmass[Npoints+Nskip-2]*pow(initradii[Npoints+Nskip-2], gammaout-3)) /
        (pow(initradii[Npoints+Nskip-1], gammaout-3) - pow(initradii[Npoints+Nskip-2], gammaout-3));
    if(totalmass<=0)  // shouldn't happen?
        totalmass = 2*initmass.back()-initmass[Npoints+Nskip-2];  // quite arbitrary
    coefmrout=(totalmass-initmass.back())*pow(rout, gammaout-3);
    
    for(int i=0; i<Npoints; i++)
    {
        logradii[i]=log(initradii[i+Nskip]);
        loginmass[i]=log((initmass[i+Nskip]-masscenter)/(totalmass-initmass[i+Nskip]));
    }
    // check if density tends to constant at r=0
    gammain=3. - (loginmass[1]-loginmass[0])/(logradii[1]-logradii[0]);
#ifdef DEBUGPRINT
    std::cerr << "Total mass = " << totalmass << ", outer density slope = " << gammaout << ", inner = " << gammain << "\n";
#endif
    //potfinite = masscenter==0 && gammain<2;
    if(!masscenter && fabs(gammain) < 0.1)
    {   // provision for higher accuracy for constant-density core (??!!)
        densconst=true;
        gammain=0;
        params.p2=initmass[Nskip+1]/initmass[Nskip]*pow(initradii[Nskip]/initradii[Nskip+1],3.0);
        params.p3=initmass[Nskip+2]/initmass[Nskip]*pow(initradii[Nskip]/initradii[Nskip+2],3.0);
        params.r1=initradii[Nskip];
        params.r2=initradii[Nskip+1];
        params.r3=initradii[Nskip+2];
        F.function = &find_alpha;
        alpha = findRoot(&F, 0.1, 5.0, 2.0);   // 2 is default for Plummer and similar types of cored potentials
        coefdrhodr = (1+alpha/3)*(1-params.p2)/(pow(initradii[2],alpha)-params.p2*pow(initradii[1],alpha));
        if(coefdrhodr<0) coefdrhodr=0;  // density should be monotonic with radius
        rho0 = initmass[Nskip]*3/(4*M_PI)/pow(initradii[Nskip],3.0)/(1 - 3./(alpha+3)*coefdrhodr*pow(initradii[Nskip],alpha));
#ifdef DEBUGPRINT
        std::cerr << "Constant-density core determined: rho(r) = " << rho0 << "*(1-" << coefdrhodr << "*r^" << alpha << ")\n";
#endif
        coefdrhodphi = -coefdrhodr*3/(4*M_PI)*alpha* pow(3/(2*M_PI*rho0), alpha/2-1);
        loginmass.push_back( 3*totalmass/(totalmass-initmass[Nskip])   // slope at r->0
            * (1-coefdrhodr*pow(initradii[Nskip],alpha))/(1-coefdrhodr*pow(initradii[Nskip],alpha)/(1+alpha/3)) );
    } else {
        densconst=false;
        loginmass.push_back((3-gammain)*(totalmass-masscenter)/(totalmass-initmass[Nskip]));  // slope at r->0
    }
    // init spline to compute mass and density within radius
    loginmass.push_back((gammaout-3)*totalmass/initmass.back());   // slope at r->infinity
    acc_loginmass = gsl_interp_accel_alloc ();
    spl_loginmass = gsl_spline_alloc (gsl_interp_ccspline, Npoints);
    gsl_spline_init (spl_loginmass, &(logradii.front()), &(loginmass.front()), Npoints);

    // inverse of the above (radius as a function of mass)
    acc_logrmass = gsl_interp_accel_alloc ();
    spl_logrmass = gsl_spline_alloc (gsl_interp_ccspline, Npoints);
    logradii.push_back(1./(3-gammain));
    logradii.push_back(1./((gammaout-3)*totalmass/initmass.back()));
    gsl_spline_init (spl_logrmass, &(loginmass.front()), &(logradii.front()), Npoints);

    // compute potential by integration of M(r)/r^2
    std::vector<double> potential(Npoints);
    F.function=&int_pot;
    for(int i=0; i<Npoints; i++)
    {
        double result, error;
        gsl_integration_qags (&F, i==0? 0 : initradii[i+Nskip-1], initradii[i+Nskip], 0, 1e-7, 1000, w, &result, &error); 
        potential[i]=result + (i>0?potential[i-1]:0);
    }
    {
        double result, error;
        gsl_integration_qagiu (&F, initradii.back(), 0, 1e-7, 1000, w, &result, &error); 
        potentialcenter = -(result+potential.back());
    }
    for(int i=0; i<Npoints; i++)
    {
        potential[i] += potentialcenter - (masscenter>0 ? masscenter/initradii[i+Nskip] : 0);
        logpotential[i]= !masscenter ? log(-1/(1/potential[i]-1/potentialcenter)) : log(-potential[i]);
    }
    potentialin=potential[0];  // store innermost point for extrapolation to r->0
    
    // init splines: potential(radius), density(potential)
    for(int i=0; i<Npoints; i++)
        logdens[i]=log(dens(initradii[i+Nskip]));

    acc_logpotential = gsl_interp_accel_alloc ();
    spl_logpotential = gsl_spline_alloc (gsl_interp_ccspline, Npoints);
    logpotential.push_back(masscenter == 0 ? -(2-gammain) :   // inner slope
        initmass[Nskip]/(potentialcenter*initradii[Nskip]-masscenter+(initmass[Nskip]-masscenter)/(2-gammain)) );
    logpotential.push_back(-1);   // outer slope (not quite true but ignore the first-order correction for extrapolation)
    gsl_spline_init (spl_logpotential, &(logradii.front()), &(logpotential.front()), Npoints);
 
    std::vector<double> logpot_reverse, logval_reverse;
    for(int i=0; i<Npoints; i++) 
        if(!gsl_isnan(logdens[Npoints-1-i])) { 
            logpot_reverse.push_back(logpotential[Npoints-1-i]);
            logval_reverse.push_back(logdens[Npoints-1-i]); 
        }
    double drhodphi_out= (gammaout-3)*coefmrout/(4*M_PI) *    // <- rho(r_out)*r_out^gamma_out
        -gammaout * pow(-pot(rout), gammaout-1) * pow(initmass.back(), -gammaout);
    logval_reverse.push_back( exp(-logdens.back()) * potential.back()*(!masscenter ? 1-potential.back()/potentialcenter : 1) * drhodphi_out);
    double drhodphi_in = densconst ?
        coefdrhodphi*pow(pot(initradii[Nskip])-potentialcenter, alpha/2-1) :
        -gammain*(3-gammain)/(4*M_PI*pow_2(initradii[Nskip])) * (1-masscenter/initmass[Nskip]);
    logval_reverse.push_back( exp(-logdens.front()) * potential.front()*(!masscenter ? 1-potential.front()/potentialcenter : 1) * drhodphi_in);
    acc_logdenspot = gsl_interp_accel_alloc ();
    spl_logdenspot = gsl_spline_alloc (gsl_interp_ccspline, logpot_reverse.size());
    gsl_spline_init (spl_logdenspot, &(logpot_reverse.front()), &(logval_reverse.front()), logpot_reverse.size());

    // compute integrated DF using Eddington inversion formula with first derivative of rho(Phi)
    F.function=&int_distr1;
    for(int i=0; i<Npoints; i++)
    {
        params.E=potential[i];
        double result, error;
        gsl_integration_qags (&F, params.E, 0, 0, 1e-10, 1000, w, &result, &error);
        result *= -(sqrt(0.125)/M_PI/M_PI);
        logintdistrf[i]=result>0 ? log(result) : -42.0;  // strange, eh?
        if(result<=0) errorcode=MS_DF_NEGATIVE;  // "Warning, f(E) not positive!\n";
    }

    // init spline for integral of f(E)
    logpot_reverse.clear(); logval_reverse.clear();
    for(int i=0; i<Npoints; i++) 
        if(!gsl_isnan(logintdistrf[Npoints-1-i])) { 
            logpot_reverse.push_back(logpotential[Npoints-1-i]);
            logval_reverse.push_back(logintdistrf[Npoints-1-i]); 
        }
    acc_logintdistrf = gsl_interp_accel_alloc ();
    spl_logintdistrf = gsl_spline_alloc (gsl_interp_cspline, logpot_reverse.size());
    gsl_spline_init (spl_logintdistrf, &(logpot_reverse.front()), &(logval_reverse.front()), logpot_reverse.size());

    // compute integrated DF using Eddington inversion formula with second derivative of rho(Phi)
    // actually this one is used by default
    F.function=&int_distr2;
    std::vector<double> logphinz, logfnz;  // values for spline initialization, but only for input points with f>0
    for(int i=Npoints-1; i>=0; i--)
    {
        params.E=potential[i];
        double result, error;
        gsl_integration_qags (&F, params.E, 0, 0, 1e-10, 1000, w, &result, &error); 
        result *= (sqrt(0.125)/M_PI/M_PI);
        if(result>0) {
            //logphinz.push_back(i>=0&&i<Npoints ? logpotential[i] :  !masscenter ? log(-1/(1/params.E-1/potentialcenter)) : log(-params.E));
            logphinz.push_back(logpotential[i]);
            logfnz.push_back(log(result));
        }
        else errorcode=MS_DF_NEGATIVE;
    }
    if(logfnz.size()<2) {  // pathological case, just fill in with zeroes
        logfnz.assign(2,0);
        logphinz.resize(2); 
        logphinz.front()=logpotential.back(); 
        logphinz.back()=logpotential.front(); 
    }
    // compute endpoint derivatives for clamped spline initialization
    {
        double result, result1, error;
        params.E=pot(rin/5);   // inward extrapolation (r->0)
        gsl_integration_qags (&F, params.E, potential[0], 0, 1e-10, 1000, w, &result, &error); 
        gsl_integration_qags (&F, potential[0], 0, 0, 1e-10, 1000, w, &result1, &error);
        result+=result1;    // for better accuracy, split interval into two
        result *= (sqrt(0.125)/M_PI/M_PI);
        double logphi_out= !masscenter ? log(-1/(1/params.E-1/potentialcenter)) : log(-params.E);
        double deriv_out = result>0 ? (log(result)-logfnz.back())/(logphi_out-logphinz.back()) : -1; // derivative at E->Phi(0), or some default value
        params.E=pot(rout*5);   // extrapolation to E->0
        gsl_integration_qags (&F, params.E, 0, 0, 1e-10, 1000, w, &result, &error); 
        result *= (sqrt(0.125)/M_PI/M_PI);
        double logphi_in= !masscenter ? log(-1/(1/params.E-1/potentialcenter)) : log(-params.E);
        double deriv_in = (log(result)-logfnz.front())/(logphi_in-logphinz.front());  // derivative at E->0
        logfnz.push_back(deriv_in);
        logfnz.push_back(deriv_out);
    }
    // init spline for f(E)
    spl_logdistrf = gsl_spline_alloc (gsl_interp_ccspline, logphinz.size());
    gsl_spline_init (spl_logdistrf, &(logphinz.front()), &(logfnz.front()), logphinz.size());

    // compute differential energy distribution  g(E)
    F.function=&int_ded;
    for(int i=0; i<Npoints; i++)
    {
        params.E=potential[i];
        double result, error;
        gsl_integration_qags (&F, 0, initradii[i+Nskip], 0, 1e-7, 1000, w, &result, &error); 
        logded[i]=log(result*16*M_PI*M_PI);
    }

    // init spline for g(E)
    logpot_reverse.clear(); logval_reverse.clear();
    for(int i=0; i<Npoints; i++) 
        if(!gsl_isnan(logded[Npoints-1-i])) { 
            logpot_reverse.push_back(logpotential[Npoints-1-i]);
            logval_reverse.push_back(logded[Npoints-1-i]); 
        }
    acc_logded = gsl_interp_accel_alloc ();
    spl_logded = gsl_spline_alloc (gsl_interp_cspline, logpot_reverse.size());
    gsl_spline_init (spl_logded, &(logpot_reverse.front()), &(logval_reverse.front()), logpot_reverse.size());

    // compute radial period as a function of energy
    F.function=&int_trad;
    for(int i=0; i<Npoints; i++)
    {
        params.E=potential[i];
        double result, error;
        gsl_integration_qags (&F, 0, initradii[i+Nskip], 0, 1e-7, 1000, w, &result, &error); 
        logtrad[i]=log(2*result);
    }

    // init spline for T_rad(E)
    logpot_reverse.clear(); logval_reverse.clear();
    for(int i=0; i<Npoints; i++) 
        if(!gsl_isnan(logtrad[Npoints-1-i])) { 
            logpot_reverse.push_back(logpotential[Npoints-1-i]);
            logval_reverse.push_back(logtrad[Npoints-1-i]); 
        }
    acc_logtrad = gsl_interp_accel_alloc ();
    spl_logtrad = gsl_spline_alloc (gsl_interp_cspline, logpot_reverse.size());
    gsl_spline_init (spl_logtrad, &(logpot_reverse.front()), &(logval_reverse.front()), logpot_reverse.size());

    // compute radius of circular orbit as a function of energy
    F.function=&find_rcirc;
    for(int i=0; i<Npoints; i++)
    {
        params.E=potential[i];
        double rlow=initradii[i+Nskip]/4, rupp=initradii[i+Nskip], result;
        while(find_rcirc(rlow, &params)>0) rlow/=2; //probably not necessary..
        result = findRoot(&F, rlow, rupp);
        logrcirc[i]=log(result);
    }

    // init spline for r_circ(E)
    logpot_reverse.clear(); logval_reverse.clear();
    for(int i=0; i<Npoints; i++) 
        if(!gsl_isnan(logrcirc[Npoints-1-i])) { 
            logpot_reverse.push_back(logpotential[Npoints-1-i]);
            logval_reverse.push_back(logrcirc[Npoints-1-i]); 
        }
    acc_logrcirc = gsl_interp_accel_alloc ();
    spl_logrcirc = gsl_spline_alloc (gsl_interp_cspline, logpot_reverse.size());
    gsl_spline_init (spl_logrcirc, &(logpot_reverse.front()), &(logval_reverse.front()), logpot_reverse.size());

    gsl_integration_workspace_free (w);
}
CMassModel::~CMassModel()
{
    gsl_spline_free(spl_loginmass);
    gsl_spline_free(spl_logrmass);
    gsl_spline_free(spl_logpotential);
    gsl_spline_free(spl_logdenspot);
    gsl_spline_free(spl_logdistrf);
    gsl_spline_free(spl_logintdistrf);
    gsl_spline_free(spl_logded);
    gsl_spline_free(spl_logtrad);
    gsl_spline_free(spl_logrcirc);
    gsl_interp_accel_free(acc_loginmass);
    gsl_interp_accel_free(acc_logrmass);
    gsl_interp_accel_free(acc_logpotential);
    gsl_interp_accel_free(acc_logdenspot);
    gsl_interp_accel_free(acc_logintdistrf);
    gsl_interp_accel_free(acc_logdistrf);
    gsl_interp_accel_free(acc_logded);
    gsl_interp_accel_free(acc_logtrad);
    gsl_interp_accel_free(acc_logrcirc);
}

}