/** \file    massmodel.h
    \brief   Class for dealing with spherical isotropic mass models
    \author  EV
    \date    2010-2013
*/
#pragma once
#include <vector>
#include <gsl/gsl_spline.h>

namespace smile{

/** class defining a spherical mass model.
    These models are constructed from an array of { r, m(r) } pairs;
    the dependence of mass, potential, distribution function, and other quantities on radius or energy 
    is pre-computed in the constructor and stored as a spline function of (logarithmically scaled) 
    radius or energy, with a careful extrapolation beyond the definition region of the spline.
    The isotropic distribution function is computed by the Eddington inversion formula.
    These models may be used independently of the main SMILE machinery, 
    or for generating initial conditions for the orbit library in COrbitLibrary::prepareInitialConditions
*/
class CMassModel{
public:
    /// internal status or error code of the model
enum MODELSTATUS {  
    MS_OK,             ///< everything is fine
    MS_ARRAY_NOT_EQ,   ///< input arrays are of different lengths
    MS_ARRAY_SHORT,    ///< input arrays are too short (less than 5 radial points)
    MS_RADII_NOT_ASC,  ///< radii are not in ascending order
    MS_MASS_NOT_ASC,   ///< mass is not increasing with radius
    MS_DF_NEGATIVE     ///< distribution function is negative somewhere (not a critical error)
};
    /// create the model from an array of { r, m(r) } pairs and pre-computes almost all necessary quantities;
    /// return errorcode
    CMassModel(const std::vector<double> &initradii, const std::vector<double> &initmas);
    ~CMassModel();
    MODELSTATUS getStatus() const    ///< return the status code set in the constructor
    { return errorcode; };
    double mass(double rad) const;    ///< return enclosed mass at given radius
    double rad(double mass) const;    ///< find radius for enclosed mass
    double dens(double rad) const;    ///< density at given radius
    double pot(double rad) const;     ///< potential at given radius
    double drhodphi(double E) const;  ///< d(rho)/d(phi) at given energy E=phi
    double d2rhodphi2(double E) const;///< d^2(rho)/d(phi)^2 at given energy E=phi
    double distrf1(double E) const;   ///< distribution function of (negative) energy, computed via differentiation of spline initialized from integral with d(rho)/d(Phi)
    double distrf2(double E) const;   ///< distribution function of (negative) energy, computed directly from integral with d^2(rho)/d(Phi)^2
    double distrf(double E) const;    ///< distribution function of (negative) energy, spline approximation to distrf2
    double ded(double E) const;       ///< differential energy distribution  g(E)
    double masse(double E) const;     ///< mass of particles having energies <E
    double trad(double E) const;      ///< period of radial orbit at energy E
    double rmax(double E) const;      ///< max radius at given energy
    double rcirc(double E) const;     ///< radius of circular orbit at energy E
    double lcirc(double E) const;     ///< ang.momentum at circular orbit
    double veldisp(double rad) const; ///< velocity dispersion at given radius
    double densproj(double rad) const;///< projected density at given projected radius
    double losvd(double rad) const;   ///< line-of-sight velocity dispersion at a given projected radius

    /// compute diffusion coefs in the Fokker-Planck equation, which depend on E and R.
    /// \param[out] DEE,DRR,DLL contain the computed coefs; if a pointer is NULL, this coef is not computed
    void DifCoefAvER(double E, double R, double *DEE, double *DRR, double *DLL) const;
    /// compute R-averaged diffusion coefs depending on E; if a pointer is NULL, this coef is not computed
    void DifCoefAvE (double E, double *DEE, double *DRR, double *DLL=NULL, double *DEL=NULL) const;
    double DRRrad(double E) const;    ///< compute limiting value of the diffusion coef: lim(DRR/R), R->0
    double samplev(double r) const;   ///< sample a velocity from distribution function at a given radius

    /// tell whether the model has no BH and a constant-density core, and if yes, give its parameters
    bool hasDensityCore(double *rho0_, double *drhodr_, double *alpha_) const {
        if(densconst) { 
            if(rho0_)   *rho0_=rho0; 
            if(drhodr_) *drhodr_=coefdrhodr; 
            if(alpha_)  *alpha_=alpha; 
        }
        return densconst;
    }
    /// return total mass of the model
    double getTotalMass() const { return totalmass; }
private:
    MODELSTATUS errorcode;     ///< if not MS_OK, signal an error in the constructor
    gsl_spline       *spl_loginmass, *spl_logrmass, *spl_logpotential, *spl_logdenspot, *spl_logintdistrf, *spl_logdistrf, *spl_logded, *spl_logtrad, *spl_logrcirc;
    gsl_interp_accel *acc_loginmass, *acc_logrmass, *acc_logpotential, *acc_logdenspot, *acc_logintdistrf, *acc_logdistrf, *acc_logded, *acc_logtrad, *acc_logrcirc;
    double potentialcenter;    ///< central potential not accounting for SMBH
    double totalmass;          ///< total mass at infinity
    double masscenter;         ///< BH mass if it exists; if masscenter!=0 then potential is divergent at origin
    double rin, rout;          ///< inner and outer grid radii
    double gammain, gammaout;  ///< power-law slopes of density at r->0 and r->infinity
    double coefmrout;          ///< extrapolation for M(r) at large radii: M=totalmass-coefmrout*r^(3-gammaout)
    double potentialin;        ///< total potential(including BH) at r=rin
    bool densconst;            ///< precautions for constant-density cores:
    double alpha, rho0, coefdrhodr, coefdrhodphi;  ///< coefficients for extrapolation to r->0 for constant-density core
};

};  // namespace