/** \file   mkspher.cpp
    \brief  create spherical mass models from a text table
    \author EV
    \date   2010-2013

This program is a simplified version of ``mkspherical'' utility from SMILE toolbox, 
intended to create spherical mass models from a text table with r, m(r) values specifying 
enclosed mass as a function of radius. 

The input is a text file with >=5 (preferrably 10-20) lines with two columns: r, m(r);
both must be increasing with line number; if m(r=0) is nonzero, it is considered to be 
a supermassive black hole at the galaxy center.

The output is a text file with a number of parameters computed as spline-interpolated 
functions of radius (potential, radial/circular period, distribution function via Eddington 
inversion formula, diffusion coefficients, etc.)

*/

/// default values of command-line parameters
const char* defv[]=
{
  "in=\n        input table with r and M(r) values",
  "out=\n       output table with r,M,Phi,f(E) and other parameters",
  "rmin=0\n        output table inner nonzero radius [0 for auto-detect]",
  "rmax=0\n        output table outer radius [0 for auto-detect]",
  "npoints=100\n   output table number of points in radius [-1 to use input points]",
  0
};

#include <vector>
#include <iostream>
#include <sstream>
#include <gsl/gsl_errno.h>
#include "massmodel.h"
#include "utils.h"
#include "config.h"

/// replacement for GSL error handler that doesn't crash program
void gsl_quiet_error_handler (const char *reason, const char *file, int line, int er)
{
    if(er == GSL_ETOL || er == GSL_EROUND || er == GSL_ESING || er == GSL_EDIVERGE || er == GSL_EMAXITER) return;  // do nothing
    std::cerr << "GSL error " << er << " in " << file << " at line " << line << ": " << reason << "\n";
}

/// main program begins here
int main(int argc, char* argv[])
{
    gsl_set_error_handler(&gsl_quiet_error_handler);
    initparam(argv, defv);
    std::string inputtab = getparam("in");
    std::string outputtab= getparam("out");
    double gridrmin = getdparam("rmin");
    double gridrmax = getdparam("rmax");
    int numNodes=getiparam("npoints");
    if(inputtab=="" || outputtab=="") {
        std::cerr << "Usage:  mkspher  in=input_text_table out=output_text_table [rmin=?] [rmax=?] [npoints=100]\n";
        return 0;
    }

    // r, M(r) table to be initialized either directly from a text file or from an N-body snapshot
    std::vector<double> radii, inmass;
    //if(inputtab!="")
    {
        // read input table
        std::ifstream in(inputtab.c_str());
        if(!in) {
            std::cerr << "Invalid input table filename: " << inputtab << "\n";
            return 0;
        }
        while(in)   // parse input file
        {
            std::string str;
            std::getline(in, str);
            if(in && str.size()>2 && str[0]!='#' && str[0]!='/' && str[0]!='%') 
            {
                std::istringstream stm(str);
                double r=0, m=0;
                stm >> r >> m;
                radii.push_back(r);
                inmass.push_back(m);
            }
        }
    }
    smile::CMassModel model(radii, inmass);
    switch(model.getStatus())
    {
    case smile::CMassModel::MS_ARRAY_NOT_EQ:  std::cerr << "Invalid format of input table file (columns of unequal length)\n"; return 0;
    case smile::CMassModel::MS_ARRAY_SHORT:   std::cerr << "Too few points\n"; return 0; 
    case smile::CMassModel::MS_RADII_NOT_ASC: std::cerr << "Radii not in ascending order!\n"; return 0;
    case smile::CMassModel::MS_MASS_NOT_ASC:  std::cerr << "Mass not increasing with radius!\n"; return 0;
    case smile::CMassModel::MS_DF_NEGATIVE:   std::cerr << "Warning, f(E) not positive!\n"; break; // no fail
    }
    double rho0, drhodr, alpha;
    if(model.hasDensityCore(&rho0, &drhodr, &alpha))
        std::cerr << "Constant-density core determined: rho(r) = " << rho0 << "*(1-" << drhodr << "*r^" << alpha << ")\n";

    //if(outputtab!="")   // output table with all computed quantities vs radius
    {
        std::ofstream out(outputtab.c_str(), std::ios_base::out);
        if(!out) {
            std::cerr << "Cannot write output table: " << outputtab << "\n";
            return 0;
        }
        std::vector<double> gridRadius;
        if(numNodes<0) {  // use input points, excluding r=0
            std::vector<double>::iterator beg=radii.begin();
            while(beg!=radii.end() && *beg<=0) beg++;
            gridRadius.assign(beg, radii.end());
            numNodes=static_cast<int>(gridRadius.size());
        } else {  // create a log-grid with reasonable inner and outer radii
            double zeroMass=model.mass(0), totalMass=model.getTotalMass()-zeroMass;
            if(gridrmin==0) {
                gridrmin=0.125;
                while(gridrmin>1e-6 && model.mass(gridrmin)-zeroMass>totalMass*1e-6) gridrmin/=2;  // a quick default algo
            }
            if(gridrmax==0) {
                gridrmax=8;
                while(gridrmax<1e6 && model.mass(gridrmax)-zeroMass<totalMass*(1-1e-3)) gridrmax*=2;
            }
            if(numNodes==0) numNodes=100;
            smile::createNonuniformGrid(numNodes, gridrmin, gridrmax, false, &gridRadius);
        }
        out << "#r\tM(r)\tPhi=E\trho\tf(E)\tg(E)\tM(E)\tperiod(E)\tr_circ(E)\tL_circ(E)\tsigma\tsigma_los\tdens_proj\tD_EE\tD_RR\tD_RR/R@R=0\n";
        out << 0 << "\t" << model.mass(0) << "\t" << model.pot(0) << "\n";
        for(int i=0; i<numNodes; i++)
        {
            if(i%10==0) std::cerr << (i*100/numNodes) << "%\r";
            double rad = gridRadius[i];
            double pot = model.pot(rad);
            double DEE, DRR;
            model.DifCoefAvE(pot, &DEE, &DRR, NULL, NULL);
            out << rad << "\t" << smile::convertToString(model.mass(rad),12) << "\t" 
                << smile::convertToString(pot,10) << "\t" << smile::convertToString(model.dens(rad),10) << "\t" 
                << model.distrf(pot) << "\t" << model.ded(pot) << "\t" << model.masse(pot) << "\t"
                << model.trad(pot) << "\t" << model.rcirc(pot) << "\t" << model.lcirc(pot) << "\t"
                << model.veldisp(rad) << "\t" << model.losvd(rad) << "\t" << model.densproj(rad) << "\t"
                << DEE << "\t" << DRR << "\t" << model.DRRrad(pot);
            out << "\n";
        }
        std::cerr << "100%\n";
        out.close();
    }

    finiparam();
    return 0;
}
