#include "utils.h"
#include <iostream>
#include <cmath>
#include <cassert>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit.h>

namespace smile{

inline double pow_2(const double x) { return x*x; }

/* --------- wrappers for GSL equation solvers and minimizers ------------- */
// ordinary 1d equation solver
double findRoot(gsl_function *F, double xlower, double xupper, double xdefault, double reltoler)
{
    if(F->function(xlower, F->params) * F->function(xupper, F->params) > 0)
        return xdefault;   // endpoints do not enclose root
    gsl_root_fsolver *solv = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
    gsl_root_fsolver_set(solv, F, xlower, xupper);
    int status=0, iter=0;
    double xroot;
    do{
        iter++;
        status = gsl_root_fsolver_iterate(solv);
        xroot  = gsl_root_fsolver_root(solv);
        xlower = gsl_root_fsolver_x_lower(solv);
        xupper = gsl_root_fsolver_x_upper(solv);
        status = gsl_root_test_interval (xlower, xupper, 0, reltoler);
    }
    while(status == GSL_CONTINUE && iter < 50);
    gsl_root_fsolver_free(solv);
    return xroot;
}

// 1d minimizer
double findMin(gsl_function *F, double xlower, double xupper, double xinit, double reltoler)
{
    gsl_min_fminimizer *minser = gsl_min_fminimizer_alloc(gsl_min_fminimizer_brent);
    double xroot=xinit;
    if(gsl_min_fminimizer_set(minser, F, xinit, xlower, xupper) == GSL_SUCCESS) 
    {
        int status=0, iter=0;
        do{
            iter++;
            status = gsl_min_fminimizer_iterate (minser);
            xroot  = gsl_min_fminimizer_x_minimum (minser);
            xlower = gsl_min_fminimizer_x_lower (minser);
            xupper = gsl_min_fminimizer_x_upper (minser);
            status = gsl_root_test_interval (xlower, xupper, 0, reltoler);
        }
        while (status == GSL_CONTINUE && iter < 50);
    }
    gsl_min_fminimizer_free(minser);
    return xroot;
}

// 1d minimizer without prior knowledge of minimum location
double findMinGuess(gsl_function *F, double xlower, double xupper, double reltoler)
{
    double xinit =(xlower+xupper)/2;
    double ylower=F->function(xlower, F->params);
    double yupper=F->function(xupper, F->params);
    double yinit =F->function(xinit, F->params);
    int iter=0;
    while( (yinit>ylower || yinit>yupper) && iter<50 && gsl_root_test_interval(xlower, xupper, 0, reltoler)==GSL_CONTINUE)
    {
        if(yinit<ylower) {
            xlower=xinit;
            ylower=yinit;
        } else 
        if(yinit<yupper) {
            xupper=xinit;
            yupper=yinit;
        } else {  // pathological case - initial guess was higher than both ends
            double xmin1=findMinGuess(F, xlower, xinit, reltoler);
            double xmin2=findMinGuess(F, xinit, xupper, reltoler);
            double ymin1=F->function(xmin1, F->params);
            double ymin2=F->function(xmin2, F->params);
            return ymin1<ymin2 ? xmin1 : xmin2;
        }
        xinit=(xlower+xupper)/2;
        yinit=F->function(xinit, F->params);
        iter++;
    }
    if(yinit<=ylower && yinit<=yupper)
        return findMin(F, xlower, xupper, xinit, reltoler);  // normal min-search
    else if(ylower<yinit) return xlower;
    else return xupper;    // otherwise return one of endpoints
}

/* ----------- generation of unequally spaced grids --------------- */
// Creation of grid with exponentially increasing cells
static double findA(double A, void* params)
{
    double dynrange = ((double*)params)[0];
    double nnodes   = ((double*)params)[1];
    if(A==0) return nnodes-dynrange;
    return (exp(A*nnodes)-1)/(exp(A)-1) - dynrange;
}
void createNonuniformGrid(size_t nnodes, double xmin, double xmax, bool zeroelem, std::vector<double> *grid)
{   // create grid so that x_k = B*(exp(A*k)-1)
    assert(grid);
    double A, B, dynrange=xmax/xmin;
    grid->resize(nnodes);
    int indexstart=zeroelem?1:0;
    if(zeroelem) { grid->at(0)=0; nnodes--; }
    if(gsl_fcmp(static_cast<double>(nnodes), dynrange, 1e-6)==0)  // no need for non-uniform grid
    {
        for(size_t i=0; i<nnodes; i++)
            grid->at(i+indexstart) = xmin+(xmax-xmin)*i/(nnodes-1);
        return;
    }
    // solve for A:  dynrange = (exp(A*nnodes)-1)/(exp(A)-1)
    double params[2] = {dynrange, static_cast<double>(nnodes)};
    gsl_function F;
    F.function=&findA;
    F.params=&params;
    // first localize the root coarsely, to avoid overflows in root solver
    double Amin=0, Amax=0;
    double step=1; while(step>10./nnodes) step/=2;
    if(dynrange>nnodes) {
        while(Amax<10 && findA(Amax, &params)<=0) Amax+=step;
        Amin=Amax-step;
    } else {
        while(Amin>-10 && findA(Amin, &params)>=0) Amin-=step;
        Amax=Amin+step;
    }
    A = findRoot(&F, Amin, Amax);
    B = xmin / (exp(A)-1);
    for(size_t i=0; i<nnodes; i++)
        grid->at(i+indexstart) = B*(exp(A*(i+1))-1);
    grid->at(nnodes-1+indexstart)=xmax;
}

/* ----------- spline functions --------------- */

// clamped cubic spline, with given values of first derivative at endpoints and linear extrapolation beyond the definition range
// Hacking into GSL spline machinery by defining a new spline type which differs from natural cubic spline only in initialization
/// \cond INTERNAL_DOCS
typedef struct
{
  double * c;
  double * g;
  double * diag;
  double * offdiag;
} cspline_state_t;
/// \endcond

// 1st derivatives at left- and rightmost points are given in ya[n] & ya[n+1]
// if one wants to have a 'natural' spline boundary condition then pass NaN as the value of derivative.
int ccspline_init (void * vstate, const double xa[], const double ya[], size_t size)
{
  cspline_state_t *state = (cspline_state_t *) vstate;

  size_t i;
  size_t num_points = size;
  size_t max_index = num_points - 1;  /* Engeln-Mullges + Uhlig "n" */
  size_t sys_size = max_index - 1;    /* linear system is sys_size x sys_size */

  double der1=ya[size], der2=ya[size+1];  // numbers must be supplied!
  for (i = 0; i < sys_size; i++)
    {
      const double h_i   = xa[i + 1] - xa[i];
      const double h_ip1 = xa[i + 2] - xa[i + 1];
      const double ydiff_i   = ya[i + 1] - ya[i];
      const double ydiff_ip1 = ya[i + 2] - ya[i + 1];
      const double g_i = (h_i != 0.0) ? 1.0 / h_i : 0.0;
      const double g_ip1 = (h_ip1 != 0.0) ? 1.0 / h_ip1 : 0.0;
      state->offdiag[i] = h_ip1;
      state->diag[i] = 2.0 * (h_ip1 + h_i);
      state->g[i] = 3.0 * (ydiff_ip1 * g_ip1 -  ydiff_i * g_i);
      if(i == 0 && !gsl_isnan(der1)) {
        state->diag[i] = 1.5 * h_i + 2.0 * h_ip1;
        state->g[i] = 3.0 * (ydiff_ip1 * g_ip1 - 1.5 * ydiff_i * g_i + 0.5 * der1);
      }
      if(i == sys_size-1 && !gsl_isnan(der2)) {
        state->diag[i] = 1.5 * h_ip1 + 2.0 * h_i;
        state->g[i] = 3.0 * (1.5 * ydiff_ip1 * g_ip1 - 0.5 * der2 - ydiff_i * g_i);
      }
    }

  if (sys_size == 1)
    {
      state->c[1] = state->g[0] / state->diag[0];
      return GSL_SUCCESS;
    }
  else
    {
      gsl_vector_view g_vec = gsl_vector_view_array(state->g, sys_size);
      gsl_vector_view diag_vec = gsl_vector_view_array(state->diag, sys_size);
      gsl_vector_view offdiag_vec = gsl_vector_view_array(state->offdiag, sys_size - 1);
      gsl_vector_view solution_vec = gsl_vector_view_array ((state->c) + 1, sys_size);
      
      int status = gsl_linalg_solve_symm_tridiag(&diag_vec.vector, &offdiag_vec.vector, &g_vec.vector, &solution_vec.vector);
      if(!gsl_isnan(der1)) 
          state->c[0] = ( 3.0*(ya[1]-ya[0])/(xa[1]>xa[0] ? xa[1]-xa[0] : 1) - 3.0*der1 - state->c[1]*(xa[1]-xa[0]) )*0.5/(xa[1]>xa[0] ? xa[1]-xa[0] : 1);
      if(!gsl_isnan(der2)) 
          state->c[max_index] = -( 3*(ya[max_index]-ya[max_index-1])/(xa[max_index]-xa[max_index-1]) - 3*der2 + state->c[max_index-1]*(xa[max_index]-xa[max_index-1]) )*0.5/(xa[max_index]-xa[max_index-1]);
      return status;
    }
}

// modified functions for spline evaluation that extrapolate last segment linearly beyond the definition region
double gsl_spline_evalx(const gsl_spline* spline, double x, gsl_interp_accel * a) 
{
    const cspline_state_t *state = (const cspline_state_t *)spline->interp->state; 
    const size_t size=spline->interp->size;
    if(x<spline->interp->xmin) {
        double dx  =  spline->x[1]-spline->x[0];
        double der = (spline->y[1]-spline->y[0])/dx - dx*(state->c[1]+2*state->c[0])/3.0;
        return spline->y[0] + der*(x-spline->interp->xmin);
    }
    else if(x>spline->interp->xmax) {
        double dx  =  spline->x[size-1]-spline->x[size-2];
        double der = (spline->y[size-1]-spline->y[size-2])/dx + dx*(state->c[size-2]+2*state->c[size-1])/3.0;
        return spline->y[size-1] + der*(x-spline->interp->xmax);
    }
    else // use standard method
        return gsl_spline_eval(spline, x, a);
}

double gsl_spline_eval_derivx(const gsl_spline* spline, double x, gsl_interp_accel * a) 
{
    const size_t size=spline->interp->size;
    const cspline_state_t *state = (const cspline_state_t *)spline->interp->state; 
    if(x<spline->interp->xmin) {
        double dx = spline->x[1]-spline->x[0];
        return (spline->y[1]-spline->y[0])/dx - dx*(state->c[1]+2*state->c[0])/3.0;
    }
    else if(x>spline->interp->xmax) {
        double dx = spline->x[size-1]-spline->x[size-2];
        return (spline->y[size-1]-spline->y[size-2])/dx + dx*(state->c[size-2]+2*state->c[size-1])/3.0;
    }
    else // use standard method
        return gsl_spline_eval_deriv(spline, x, a);
}

double gsl_spline_eval_deriv2x(const gsl_spline* spline, double x, gsl_interp_accel * a) 
{
    if(x<spline->interp->xmin || x>spline->interp->xmax) 
        return 0;
    else // use standard method
        return gsl_spline_eval_deriv2(spline, x, a);
}

const gsl_interp_type ccspline_type = 
{
  "ccspline", 
  3,
  gsl_interp_cspline->alloc,
  &ccspline_init,
  gsl_interp_cspline->eval,
  gsl_interp_cspline->eval_deriv, 
  gsl_interp_cspline->eval_deriv2, 
  gsl_interp_cspline->eval_integ,
  gsl_interp_cspline->free
};

const gsl_interp_type * gsl_interp_ccspline = &ccspline_type;

}  // namespace